package com.tocea.easycoverage.framework;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.tocea.easycoverage.framework.api.IEasycoverageChecker;
import com.tocea.easycoverage.framework.api.IInstanceProvider;
import com.tocea.easycoverage.framework.exception.EasycoverageAssertionError;
import com.tocea.easycoverage.framework.providers.DefaultInstanceProvider;
import com.tocea.easycoverage.framework.utils.Utils;

/**
 * 
 * A helper class to run checks on to build other runners on top of it. It takes
 * an instance provider, a list of classes, a list of class checkers and a list
 * of method checkers. getCheckers returns a set of class or method checkers
 * (depending on the parameter), while runChecks will run all the possible
 * checks and print the results in System.out.
 * 
 * @author Romain Delamare
 * 
 */
public class CheckersRunner {

	private static final Logger							LOGGER			= LoggerFactory
																				.getLogger(CheckersRunner.class);

	private IInstanceProvider							provider;
	private final Set<Class<?>>							classes			= new HashSet<Class<?>>();
	private final Set<Class<? extends ClassChecker>>	classCheckers	= new HashSet<Class<? extends ClassChecker>>();
	private final Set<Class<? extends MethodChecker>>	methodCheckers	= new HashSet<Class<? extends MethodChecker>>();
	private int											timeout			= 0;
	private final ExecutorService						executor		= Executors
																				.newCachedThreadPool();

	/**
	 * <p>
	 * Runs checks on all the classes previously specified. For each class it
	 * will run all the class checkers that can be applied, and for each method
	 * it will run all the method checkers that can be applied.
	 * </p>
	 * 
	 * <p>
	 * Prints the result in System.out.
	 * </p>
	 */
	public void runChecks() {
		final PrintStream out = System.out;
		System.setOut(new PrintStream(new ByteArrayOutputStream()));
		int runChecks = 0;
		final int failedChecks = 0;
		for (final Class<?> clazz : getClasses()) {
			out.println("****** " + clazz.getName() + " ******");
			final Set<ClassChecker> classCheckers = getCheckers(clazz);
			runChecks += classCheckers.size();
			for (final ClassChecker checker : classCheckers)
				runCheck(checker, out, false);
			for (final Method method : Utils.getDeclaredPublicMethods(clazz)) {
				final Set<MethodChecker> methodCheckers = getCheckers(method);
				if (methodCheckers.isEmpty())
					continue;
				out.println("\t****** " + Utils.getQualifiedName(method)
						+ " ******");
				runChecks += methodCheckers.size();
				for (final MethodChecker checker : methodCheckers)
					runCheck(checker, out, true);
				out.print("\t*******");
				for (int i = 0; i < Utils.getQualifiedName(method).length(); i++)
					out.print("*");
				out.println("*******");
			}
			out.print("*******");
			for (int i = 0; i < clazz.getName().length(); i++)
				out.print("*");
			out.println("*******");
		}
		out.println("\nRun Checks: " + runChecks + ", Failed: " + failedChecks);
		System.setOut(out);
	}

	private boolean runCheck(final IEasycoverageChecker checker,
			final PrintStream out, final boolean indent) {
		boolean failed = false;
		final Future<?> future = executor.submit(new CheckerTask(checker));
		try {
			if (timeout > 0)
				future.get(timeout, TimeUnit.MILLISECONDS);
			else
				future.get();
		} catch (final InterruptedException e) {
			failed = true;
			out.println(indent ? "\t - " : " - "
					+ checker.getClass().getSimpleName()
					+ " failed: interrupted");
		} catch (final TimeoutException e) {
			if (future != null && !future.isDone() && !future.isCancelled())
				future.cancel(true);
			failed = true;
			out.println(indent ? "\t - " : " - "
					+ checker.getClass().getSimpleName() + " failed: timeout");
		} catch (final ExecutionException e) {
			failed = true;
			out.println(indent ? "\t - " : " - "
					+ checker.getClass().getSimpleName() + " failed: "
					+ e.getMessage());
		}
		return failed;
	}

	/**
	 * Gets all the applicable class checkers for a given class.
	 * 
	 * @param clazz
	 *            the class to check
	 * @return the set of applicable class checkers
	 */
	public Set<ClassChecker> getCheckers(final Class<?> clazz) {
		final Set<ClassChecker> checkers = new HashSet<ClassChecker>();
		if (getProvider().canProvide(clazz))
			for (final Class<? extends ClassChecker> checkerClass : getClassCheckers()) {
				try {
					final ClassChecker checker = checkerClass.getConstructor(
							IInstanceProvider.class, Class.class).newInstance(
							getProvider(), clazz);
					if (checker.shouldCheck())
						checkers.add(checker);
				} catch (final NoSuchMethodException e) {
					throw new EasycoverageAssertionError(
							checkerClass.getCanonicalName()
									+ " should have a public constructor with two arguments (IInstanceProvider, Class)");
				} catch (final Exception e) {
					LOGGER.error("Error while instantiating {}: {}",
							checkerClass, e.getMessage());
				}
			}
		return checkers;
	}

	/**
	 * Gets all the applicable method checkers for a given method.
	 * 
	 * @param method
	 *            the method to check
	 * @return the set of applicable method checkers
	 */
	public Set<MethodChecker> getCheckers(final Method method) {
		final Set<MethodChecker> checkers = new HashSet<MethodChecker>();
		if (method != null
				&& getProvider().canProvide(method.getDeclaringClass()))
			for (final Class<? extends MethodChecker> checkerClass : getMethodCheckers()) {
				try {
					final MethodChecker checker = checkerClass.getConstructor(
							IInstanceProvider.class, Method.class).newInstance(
							getProvider(), method);
					if (checker.shouldCheck())
						checkers.add(checker);
				} catch (final NoSuchMethodException e) {
					throw new EasycoverageAssertionError(
							checkerClass.getCanonicalName()
									+ " should have a public constructor with two arguments (IInstanceProvider, Class)");
				} catch (final Exception e) {
					LOGGER.error("Error while instantiating {}: {}",
							checkerClass, e.getMessage());
				}
			}
		return checkers;
	}

	public IInstanceProvider getProvider() {
		if (provider == null)
			provider = new DefaultInstanceProvider();
		return provider;
	}

	public void setProvider(final IInstanceProvider provider) {
		this.provider = provider;
	}

	public Set<Class<?>> getClasses() {
		return classes;
	}

	public void addClass(final Class<?> clazz) {
		classes.add(clazz);
	}

	public void addClasses(final Collection<Class<?>> classes) {
		if (classes != null)
			this.classes.addAll(classes);
	}

	/**
	 * @deprecated Since 0.2.3. Use {@link #addClasses(Collection)} instead.
	 */
	@Deprecated
	public void setClasses(final Set<Class<?>> classes) {
		addClasses(classes);
	}

	public Set<Class<? extends ClassChecker>> getClassCheckers() {
		return classCheckers;
	}

	public void addClassChecker(final Class<? extends ClassChecker> checker) {
		classCheckers.add(checker);
	}

	public void addClassCheckers(
			final Collection<Class<? extends ClassChecker>> checkers) {
		if (checkers != null)
			classCheckers.addAll(classCheckers);
	}

	/**
	 * @deprecated Since 0.2.3. Use {@link #addClassCheckers(Collection)}
	 *             instead.
	 */
	@Deprecated
	public void setClassCheckers(
			final Set<Class<? extends ClassChecker>> checkers) {
		addClassCheckers(checkers);
	}

	public Set<Class<? extends MethodChecker>> getMethodCheckers() {
		return methodCheckers;
	}

	public void addMethodChecker(final Class<? extends MethodChecker> checker) {
		methodCheckers.add(checker);
	}

	public void addMethodCheckers(
			final Collection<Class<? extends MethodChecker>> checkers) {
		if (checkers != null)
			methodCheckers.addAll(checkers);
	}

	/**
	 * @deprecated Since 0.2.3. Use {@link #addMethodCheckers(Collection)}
	 *             instead.
	 */
	@Deprecated
	public void setMethodCheckers(
			final Set<Class<? extends MethodChecker>> checkers) {
		addMethodCheckers(checkers);
	}

	public int getTimeout() {
		return timeout;
	}

	public void setTimeout(final int timeout) {
		this.timeout = timeout;
	}

	private static class CheckerTask implements Runnable {

		private final IEasycoverageChecker	checker;

		public CheckerTask(final IEasycoverageChecker checker) {
			this.checker = checker;
		}

		public void run() {
			checker.check();
		}

	}

}
