/**
 * Easy Coverage for Java
 * <p>
 * Copyright 2011-2012 Tocea Technology Solutions http://www.tocea.com
 * <p>
 * This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
 * <p>
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 * <p>
 * You should have received a copy of the GNU Lesser General Public License along with this library. If not, see
 * <http://www.gnu.org/licenses/>.
 * <p>
 **/

package com.tocea.easycoverage.framework;

import java.io.File;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This type describes the scan scanner.
 * <p>
 * 
 * @author $Author: sleroy $
 * @since 12 mai 2012
 * @copyright Copyright (C) 2010-2012 TOCEA
 */

public final class ClassScanner {

	private static final Logger	LOGGER	= LoggerFactory
												.getLogger(ClassScanner.class);
	private final Set<Class<?>>	classes	= new HashSet<Class<?>>();

	/**
	 * Builds a class scanner
	 */
	public ClassScanner() {

		super();
	}

	/**
	 * This method performs the addition of a set of class
	 * 
	 * @param _classes
	 *            the classes to add.
	 */
	public ClassScanner addClasses(final Class<?>... _classes) {

		if (_classes == null) {
			return this;
		}
		for (final Class<?> clazz : _classes) {
			addClass(clazz);
		}
		return this;
	}

	/**
	 * This method performs the addition of a set of class
	 * 
	 * @param _classes
	 *            the classes to add.
	 */
	public ClassScanner addClasses(final Iterable<Class<?>> _classes) {

		if (_classes == null) {
			throw new IllegalArgumentException();
		}

		for (final Class<?> clazz : _classes) {
			addClass(clazz);
		}
		return this;
	}

	/**
	 * Returns the classes field.
	 * 
	 * @return the classes
	 */
	public Set<Class<?>> getClasses() {

		return classes;
	}

	/**
	 * Removes classes that will not be tested.
	 */
	public ClassScanner removeClasses(final Class<?>... _classes) {

		if (_classes == null) {
			throw new IllegalArgumentException();
		}

		for (final Class<?> clazz : _classes) {
			LOGGER.debug("Remove the class {}", clazz);
			classes.remove(clazz);
		}
		return this;
	}

	/**
	 * This method removes all the classes matching a regexp.
	 * 
	 * @param _regexp
	 *            the regular expression
	 */
	public void removeClasses(final Pattern _regexp) {
		for (final Class<?> cl : new ArrayList<Class<?>>(classes)) {
			if (_regexp.matcher(cl.getCanonicalName()).matches()) {
				classes.remove(cl);
			}
		}

	}

	/**
	 * This method removes all the classes already initialized
	 * 
	 * @param _packageName
	 *            the string.
	 */
	public void removePackage(final String _packageName) {
		for (final Class<?> cl : new ArrayList<Class<?>>(classes)) {
			if (_packageName.equals(cl.getPackage().getName())) {
				classes.remove(cl);
			}
		}

	}

	/**
	 * This method performs the scanning of a folder to find all classes.
	 * 
	 * @param _folder
	 *            the classes to add.
	 */
	public ClassScanner scanFolder(final File _folder) {

		if (_folder == null || !_folder.exists() || !_folder.canRead()) {
			throw new IllegalArgumentException();
		}

		final LinkedList<File> queue = new LinkedList<File>();
		final String absolutePath = _folder.getAbsolutePath();
		queue.push(_folder);
		while (!queue.isEmpty()) {
			final File rootPath = queue.pop();
			final File[] listFiles = rootPath.listFiles();
			if (listFiles == null) {
				continue;
			}
			for (final File _arg0 : listFiles) {
				if (_arg0.isDirectory()) {
					queue.push(_arg0);
				}
				if (_arg0.isDirectory() || !_arg0.getName().endsWith(".java")
						|| _arg0.equals(_folder)) {
					continue;
				}
				final String absolutePath2 = _arg0.getAbsolutePath();
				final String replace = absolutePath2.substring(
						absolutePath.length() + 1).replace(File.separatorChar,
						'.');

				final String className = replace.substring(0,
						replace.length() - 5);
				try {
					addClass(Thread.currentThread().getContextClassLoader()
							.loadClass(className));
				} catch (final ClassNotFoundException e) {
					LOGGER.warn("Class not found {}", e);
				}

			}
		}

		return this;
	}

	/**
	 * This method performs the addition of a class.
	 * 
	 * @param _clazz
	 *            the class
	 */
	private void addClass(final Class<?> _clazz) {

		LOGGER.debug("Addition of the class {}", _clazz);
		classes.add(_clazz);
	}

}
